from enum import Enum

import pytest
from qcio import CalcType

from qccodec.exceptions import RegistryError
from qccodec.registry import ParserSpec


class MockFileType(Enum):
    stdout = "stdout"
    input = "input"
    directory = "directory"


def mock_parser_function():
    pass


def another_mock_parser_function():
    pass


def test_get_parsers_by_program(parser_registry):
    parser_spec = ParserSpec(
        parser=mock_parser_function,
        filetype=MockFileType.stdout,
        required=True,
        calctypes=[CalcType.energy],
        program="mock_program",
        target="energy",
    )
    parser_registry.register(parser_spec)
    parsers = parser_registry.get_parsers(program="mock_program")
    assert len(parsers) == 1
    assert parsers[0] == parser_spec


def test_get_parsers_by_program_and_filetype(parser_registry):
    parser_spec = ParserSpec(
        parser=mock_parser_function,
        filetype=MockFileType.stdout,
        required=True,
        calctypes=[CalcType.energy],
        program="mock_program",
        target="energy",
    )
    parser_registry.register(parser_spec)
    parsers = parser_registry.get_parsers(
        program="mock_program", filetype=MockFileType.stdout
    )
    assert len(parsers) == 1
    assert parsers[0] == parser_spec


def test_get_parsers_by_program_and_calctype(parser_registry):
    parser_spec = ParserSpec(
        parser=mock_parser_function,
        filetype=MockFileType.stdout,
        required=True,
        calctypes=[CalcType.energy],
        program="mock_program",
        target="energy",
    )
    parser_registry.register(parser_spec)
    parsers = parser_registry.get_parsers(
        program="mock_program", calctype=CalcType.energy
    )
    assert len(parsers) == 1
    assert parsers[0] == parser_spec


def test_get_parsers_no_match(parser_registry):
    parser_spec = ParserSpec(
        parser=mock_parser_function,
        filetype=MockFileType.stdout,
        required=True,
        calctypes=[CalcType.energy],
        program="mock_program",
        target="energy",
    )
    parser_registry.register(parser_spec)
    parsers = parser_registry.get_parsers(
        program="mock_program", filetype=MockFileType.directory
    )
    assert len(parsers) == 0


def test_supported_programs(parser_registry):
    parser_spec = ParserSpec(
        parser=mock_parser_function,
        filetype=MockFileType.stdout,
        required=True,
        calctypes=[CalcType.energy],
        program="mock_program",
        target="energy",
    )
    parser_registry.register(parser_spec)
    programs = parser_registry.supported_programs()
    assert len(programs) == 1
    assert programs[0] == "mock_program"


def test_supported_filetypes(parser_registry):
    parser_spec = ParserSpec(
        parser=mock_parser_function,
        filetype=MockFileType.stdout,
        required=True,
        calctypes=[CalcType.energy],
        program="mock_program",
        target="energy",
    )
    parser_registry.register(parser_spec)
    filetypes = parser_registry.supported_filetypes("mock_program")
    assert len(filetypes) == 1
    assert filetypes[0] == "stdout"


def test_get_spec_for_registered_parser(parser_registry):
    parser_spec = ParserSpec(
        parser=mock_parser_function,
        filetype=MockFileType.stdout,
        required=True,
        calctypes=[CalcType.energy],
        program="mock_program",
        target="energy",
    )
    parser_registry.register(parser_spec)
    spec = parser_registry.get_spec(mock_parser_function)
    assert spec == parser_spec


def test_get_spec_for_unregistered_parser_raises_error(parser_registry):
    with pytest.raises(RegistryError, match="No parser registered for"):
        parser_registry.get_spec(mock_parser_function)


def test_parser_spec_without_target_raises_error(parser_registry):
    with pytest.raises(RegistryError):
        parser_spec = ParserSpec(
            parser=mock_parser_function,
            filetype=MockFileType.stdout,
            required=True,
            calctypes=[CalcType.energy],
            program="mock_program",
            target=None,
        )


def test_directory_parser_spec_without_target_ok(parser_registry):
    ParserSpec(
        parser=mock_parser_function,
        filetype=MockFileType.directory,
        required=True,
        calctypes=[CalcType.energy],
        program="mock_program",
        target=None,
    )


def test_register_duplicate_target_raises_error(parser_registry):
    parser_spec1 = ParserSpec(
        parser=mock_parser_function,
        filetype=MockFileType.stdout,
        required=True,
        calctypes=[CalcType.energy],
        program="mock_program",
        target="energy",
    )
    parser_spec2 = ParserSpec(
        parser=another_mock_parser_function,
        filetype=MockFileType.stdout,
        required=True,
        calctypes=[CalcType.energy],
        program="mock_program",
        target="energy",
    )
    parser_registry.register(parser_spec1)
    with pytest.raises(RegistryError, match="Duplicate parser target"):
        parser_registry.register(parser_spec2)


def test_register_different_targets_same_program(parser_registry):
    parser_spec1 = ParserSpec(
        parser=mock_parser_function,
        filetype=MockFileType.stdout,
        required=True,
        calctypes=[CalcType.energy],
        program="mock_program",
        target="energy",
    )
    parser_spec2 = ParserSpec(
        parser=another_mock_parser_function,
        filetype=MockFileType.stdout,
        required=True,
        calctypes=[CalcType.gradient],
        program="mock_program",
        target="gradient",
    )
    parser_registry.register(parser_spec1)
    parser_registry.register(parser_spec2)
    assert len(parser_registry.registry["mock_program"]) == 2



import argparse
from pathlib import Path

from .codec import decode


def main():
    parser = argparse.ArgumentParser(
        description="Parse Quantum Chemistry output files into structured JSON or "
        "Python objects."
    )
    parser.add_argument("program", help="Name of the program")
    parser.add_argument(
        "calctype", help="Type of calculation. See qcio.CalcType for options"
    )
    parser.add_argument(
        "stdout",
        help="Path to the stdout file (optional)",
        nargs="?",
        default=None,
    )
    parser.add_argument(
        "directory",
        help="Path to the directory containing the output files (optional)",
        nargs="?",
        default=None,
    )
    args = parser.parse_args()

    stdout_contents = Path(args.stdout).read_text() if args.stdout else None
    results = decode(
        args.program, args.calctype, stdout=stdout_contents, directory=args.directory
    )
    print(results.model_dump_json(indent=4, exclude_unset=True))



import inspect
import shutil
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable, Optional, Union

import pytest
from qcio import CalcType, ProgramInput
from qcio.utils import water

from qccodec.codec import decode
from qccodec.exceptions import MatchNotFoundError
from qccodec.registry import ParserRegistry, registry


@pytest.fixture(scope="session")
def test_data_dir():
    """Test data directory Path"""
    return Path(__file__).parent / "data"


@pytest.fixture
def parser_registry():
    return ParserRegistry()


@pytest.fixture
def terachem_file(test_data_dir):
    """Return a function that reads a file from the 'terachem' subdirectory."""

    def _read(filename: str) -> str:
        return (test_data_dir / "terachem" / filename).read_text()

    return _read


@pytest.fixture
def crest_file(test_data_dir):
    """Return a function that reads a file from the 'crest' subdirectory."""

    def _read(filename: str) -> str:
        return (test_data_dir / "crest" / filename).read_text()

    return _read


@pytest.fixture(scope="session")
def prog_inp():
    """Create a function that returns a ProgramInput object with a specified
    calculation type."""

    def create_prog_input(calctype):
        return ProgramInput(
            structure=water,
            calctype=calctype,
            # Integration tests depend up this model; do not change
            model={"method": "hf", "basis": "sto-3g"},
            # Tests depend upon these keywords; do not change
            keywords={
                "purify": "no",
                "some-bool": False,
            },
        )

    return create_prog_input


@dataclass
class ParserTestCase:
    """Test case for a parser function.

    Attributes:
        name: A human-readable name for the test case.
        parser: The parser function to be tested.
        contents: The input data for the parser, either as a string or a Path.
        contents_stdout: Boolean indicating if the contents should be treated as stdout.
        calctype: The calculation type for the test case.
        success: Boolean indicating if the parser should succeed on the contents.
        decode_exc: Boolean indicating if an exception is expected for MatchNotFound
            errors during decode. Default is True. required=False parsers should not
            raise an error.
        answer: The expected output from the parser.
        clear_registry: Boolean indicating if the registry should be cleared of all other
            parsers before running the test. Default is True.
        extra_files: Optional list of additional files to be copied to the test directory.
        extra_files_names: Optional list of names for the extra files, these names will
            be used as the filenames in the test directory. If not provided, the
            filenames will be the same as in the extra_files list.

    """

    name: str
    parser: Callable
    contents: Union[str, Path]
    contents_stdout: bool
    calctype: CalcType
    success: bool
    decode_exc: bool = True
    answer: Optional[Any] = None
    clear_registry: bool = True
    extra_files: Optional[list[str]] = None
    extra_files_names: Optional[list[str]] = None


def _load_contents(directory, contents):
    """Load the contents of a TestCase."""
    if isinstance(contents, Path):
        # Contents is a Path, so we read the file directly.
        return (directory / contents).read_text()
    else:
        # Contents is a string, so we assume it's the content itself.
        return contents


def get_target_value(results, target):
    """Lookup the value of a target in the results dictionary.
    Args:
        results: The results dictionary.
        target: The target key or tuple of keys to look up.
    """
    keys = target if isinstance(target, tuple) else (target,)
    d = results
    for key in keys[:-1]:
        d = d.setdefault(key, {})
    return d.get(keys[-1], None)


def _test_parser_direct(tc, contents, directory, prog_inp, parser_spec):
    """Test the parser function directly with the provided contents.

    Args:
        tc: The TestCase object containing the test parameters.
        contents: The contents to be parsed.
        directory: The directory containing the test data files.
        prog_inp: The function to create ProgramInput objects.
        parser_spec: The specification of the parser being tested.
    """
    if tc.success:
        # Successful execution of directory parser
        if parser_spec.filetype == "directory":
            parsed = tc.parser(
                directory,
                contents if tc.contents_stdout else None,
                prog_inp(tc.calctype),
            )
        else:
            # Successful execution of file parser
            parsed = tc.parser(contents)
        assert parsed == tc.answer, (
            f"{tc.name}: Direct parsing returned {parsed} instead of expected {tc.answer}"
        )

    else:
        with pytest.raises(MatchNotFoundError):
            if parser_spec.filetype == "directory":
                # Expect an exception for directory parser
                tc.parser(
                    directory,
                    contents if tc.contents_stdout else None,
                    prog_inp(tc.calctype),
                )
            else:
                # Expect an exception for file parser
                tc.parser(contents)


def _test_decode_integration(tc, contents, directory, prog_inp, program, parser_spec):
    """
    Test the decode() integration, using only the parser under test (unless tc.clear_registry is False).

    Args:
        tc: The TestCase object containing the test parameters.
        contents: The contents to be parsed (needed if tc.contents_stdout is True).
        directory: The directory containing the test data files.
        prog_inp: The function to create ProgramInput objects.
        program: The name of the program being tested.
        parser_spec: The specification of the parser being tested.

    Notes:
        - This test doesn't work for cases when the parser is a directory parser
            (i.e., filetype == "directory") and then we pass fake contents like
            "No energy here" but the parser is operating on, e.g., the crest.engrad file
            in the test data directory. It will not raise an exception as the parser will
            see the crest.engrad file. We may want to update this function to use a
            temporary directory and then write the contents variable to a file in that
            directory using parser_spec.filetype.value.

    """
    prog_input = prog_inp(tc.calctype)
    if tc.success:
        # Successful execution of decode
        result = decode(
            program,
            tc.calctype,
            stdout=(contents if tc.contents_stdout else None),
            directory=directory,
            input_data=prog_input,
            as_dict=True,
        )
        if parser_spec.target is not None:
            final_value = get_target_value(result, parser_spec.target)
        else:
            # If the target is None, we assume the entire result is the value.
            final_value = result
        assert final_value == tc.answer, (
            f"{tc.name}: decode() returned {final_value} for target '{parser_spec.target}' "
            f"instead of expected {tc.answer}"
        )
    else:
        if tc.decode_exc:
            # Failed execution and required is True
            with pytest.raises(MatchNotFoundError):
                decode(
                    program,
                    tc.calctype,
                    stdout=(contents if tc.contents_stdout else None),
                    directory=directory,
                    input_data=prog_input,
                )
        else:
            # Failed execution and required is False
            result = decode(
                program,
                tc.calctype,
                stdout=(contents if tc.contents_stdout else None),
                directory=directory,
                input_data=prog_input,
                as_dict=True,
            )
            final_value = get_target_value(result, parser_spec.target)
            assert final_value in (None, {}), (
                f"{tc.name}: decode() returned non-empty value {final_value} when an empty result was expected."
            )


@contextmanager
def restore_registry(program: str):
    """Context manager that restores the original registry for a program."""
    original = registry.get_parsers(program)
    try:
        yield
    finally:
        registry.registry[program] = original


def run_test_harness(test_data_dir, prog_inp, tmp_path, tc):
    """
    Run the full test harness:
      1. Load the contents.
      2. Run the parser directly.
      3. Temporarily restrict the registry to only the parser under test,
         and run the decode() integration test.
    """
    program = inspect.getmodule(tc.parser).__name__.split(".")[-1]
    contents = _load_contents(test_data_dir / program, tc.contents)
    # Get the spec for the parser under test.
    parser_spec = registry.get_spec(tc.parser)

    # Copy over extra files if provided.
    if tc.extra_files:
        # Use the extra_files_names if provided, otherwise use extra_files.
        extra_files_names = tc.extra_files_names or tc.extra_files
        for extra_file, extra_file_name in zip(tc.extra_files, extra_files_names):
            # Copy the extra file to the temporary directory.
            shutil.copy(
                test_data_dir / program / extra_file,
                tmp_path / extra_file_name,
            )

    if not tc.contents_stdout:
        # If contents is not stdout, copy the test data files.
        if parser_spec.filetype == "directory":
            # Use the name given in the contents variable.
            filepath = tmp_path / tc.contents
        else:
            # Use the filetype from the parser spec.
            filepath = tmp_path / parser_spec.filetype.value
        # Write the contents to the temporary file.
        filepath.write_text(contents)
    # Test the parser directly.
    _test_parser_direct(tc, contents, tmp_path, prog_inp, parser_spec)

    # Now test integration via decode() with a restricted registry.
    with restore_registry(program):
        if tc.clear_registry:
            # Clear the registry of all other parsers for this program.
            registry.registry.pop(program)
            registry.registry[program] = [parser_spec]
        _test_decode_integration(tc, contents, tmp_path, prog_inp, program, parser_spec)



class BaseError(Exception):
    """Base qccodec exceptions"""


class DecoderError(BaseError):
    """Exception raised when a decoder error occurs"""


class EncoderError(BaseError):
    """Exception raised when a encoder error occurs"""


class ParserError(BaseError):
    """Base exception for parsers"""


class MatchNotFoundError(ParserError):
    """Exception raised when a parsing match is not found"""

    def __init__(self, regex: str, contents: str):
        self.regex = regex
        self.contents = contents
        super().__init__(
            f"Could not locate match for regex: '{regex}' in contents: '{contents}'"
        )


class RegistryError(BaseError):
    """Exception raised when a registry error occurs"""


class DataCollectorError(BaseError):
    """Exception raised when a data collector error occurs"""



"""Parsers for TeraChem output files."""

import re
from enum import Enum
from pathlib import Path
from typing import Generator, Optional, Union

from qcconst import constants
from qcio import (
    CalcType,
    ProgramInput,
    ProgramOutput,
    Provenance,
    SinglePointResults,
    Structure,
)

from qccodec.exceptions import MatchNotFoundError, ParserError

from ..registry import register
from .utils import re_finditer, re_search


class TeraChemFileType(str, Enum):
    """TeraChem filetypes."""

    STDOUT = "stdout"
    DIRECTORY = "directory"


def iter_files(
    stdout: Optional[str], directory: Optional[Union[Path, str]]
) -> Generator[tuple[TeraChemFileType, Union[str, bytes, Path]], None, None]:
    """
    Iterate over the files in a TeraChem output directory.

    If stdout is provided, yields a tuple for it.

    If directory is provided, iterates over the directory to yield files according to
    program-specific logic.

    Args:
        stdout: The contents of the TeraChem stdout file.
        directory: The path to the directory containing the TeraChem output files.

    Yields:
        (FileType, contents) tuples for a program's output.
    """
    if stdout is not None:
        yield TeraChemFileType.STDOUT, stdout

    if directory is not None:
        directory = Path(directory)
        # Check if the directory exists and is a directory
        if not directory.exists() or not directory.is_dir():
            raise ParserError(
                f"Directory {directory} does not exist or is not a directory."
            )
        yield TeraChemFileType.DIRECTORY, directory


@register(
    filetype=TeraChemFileType.STDOUT,
    calctypes=[CalcType.energy, CalcType.gradient, CalcType.hessian],
    target="energy",
)
def parse_energy(contents: str) -> float:
    """Parse the final energy from TeraChem stdout.

    NOTE:
        - Works on frequency files containing many energy values because re.search()
          returns the first result.
    """
    regex = r"FINAL ENERGY: (-?\d+(?:\.\d+)?)"
    return float(re_search(regex, contents).group(1))


@register(
    filetype=TeraChemFileType.STDOUT,
    calctypes=[CalcType.gradient, CalcType.hessian],
    target="gradient",
)
def parse_gradient(contents: str) -> list[list[float]]:
    """Parse the gradient from TeraChem stdout.

    Returns:
        The gradient as a list of 3-element lists.

    Raises:
        MatchNotFoundError: If no gradient data is found.

    Notes:
        - This works for exciting state gradients as well because TeraChem prints out
            the targeted gradient as the "regular" gradient.
    """
    regex = (
        r"(?<=dE\/dX\s{12}dE\/dY\s{12}dE\/dZ\n)"  # Just after the header line
        r"[\d\.\-\s]"  # the gradient block itself
        r"+(?=\n(?:--|-=|\Z))"  # top at -- (grad) or -= (opt) or end of file (when I manually split opt logs)
    )
    match = re_search(regex, contents)
    # Convert the found numbers to floats.
    values = [float(val) for val in match.group(0).split()]
    # Group the values into chunks of 3 (for x, y, z).
    gradient = [values[i : i + 3] for i in range(0, len(values), 3)]
    return gradient


@register(
    filetype=TeraChemFileType.STDOUT, calctypes=[CalcType.hessian], target="hessian"
)
def parse_hessian(contents: str) -> list[list[float]]:
    """Parse Hessian Matrix from TeraChem stdout in one pass.

    Args:
        contents: The contents of the TeraChem stdout file.

    Returns:
        A square Hessian matrix as a list of lists of floats.

    Raises:
        MatchNotFoundError: If no Hessian data is found.
        ParserError: If the extracted numbers cannot form a proper square matrix.
    """
    regex = r"\s+(?P<atom_number>\d+)\s(?P<vals>(?:\s-?\d.\d{15}e[+-]\d{2})+)"
    hessian: list[list[float]] = []

    matches = re_finditer(regex, contents)
    # Iterate over all matches and populate the Hessian matrix
    for match in matches:
        atom_index = int(match.group("atom_number")) - 1  # Convert to zero-based index
        # Check if the Hessian matrix has enough rows
        if len(hessian) <= atom_index:
            # Add empty rows if necessary
            hessian.extend([[] for _ in range(atom_index - len(hessian) + 1)])
        # Extract the values and convert them to floats
        vals = [float(val) for val in match.group("vals").split()]
        # Add values to the corresponding row
        hessian[atom_index].extend(vals)

    # Verify that the Hessian is a square matrix.
    for i, row in enumerate(hessian):
        if len(row) != len(hessian):
            raise ParserError(
                f"Hessian matrix is not square: row {i} has {len(row)} elements, expected {len(hessian)}."
            )
    return hessian


@register(filetype=TeraChemFileType.STDOUT, target=("extras", "program_version"))
def parse_version(contents: str) -> str:
    """Parse version contents plus git commit from TeraChem stdout.

    Matches format of 'terachem --version' on command line.

    Example:
        'v1.9-2022.03-dev [4daa16dd21e78d64be5415f7663c3d7c2785203c]'
    """
    return f"{parse_terachem_version(contents)} [{parse_version_control_details(contents)}]"


@register(filetype=TeraChemFileType.STDOUT, target="calcinfo_natoms")
def parse_natoms(contents: str) -> int:
    """Parse number of atoms value from TeraChem stdout.

    Returns:
        The number of atoms as an integer.

    Raises:
        MatchNotFoundError: If the regex does not match.
    """
    regex = r"Total atoms:\s*(\d+)"
    match = re_search(regex, contents)
    return int(match.group(1))


@register(filetype=TeraChemFileType.STDOUT, target="calcinfo_nmo")
def parse_nmo(contents: str) -> int:
    """Parse the number of molecular orbitals from TeraChem stdout.

    Returns:
        The number of molecular orbitals as an integer.

    Raises:
        MatchNotFoundError: If the regex does not match.
    """
    regex = r"Total orbitals:\s*(\d+)"
    match = re_search(regex, contents)
    return int(match.group(1))


@register(
    filetype=TeraChemFileType.DIRECTORY,
    calctypes=[CalcType.optimization],
    target="trajectory",
)
def parse_trajectory(
    directory: Union[Path, str],
    stdout: str,
    input_data: ProgramInput,
) -> list[ProgramOutput]:
    """Parse the output directory of a TeraChem optimization calculation into a trajectory.

    Args:
        directory: Path to the directory containing the TeraChem output files.
        stdout: The contents of the TeraChem stdout file.
        input_data: The input object used for the calculation.

    Returns:
        A list of ProgramOutput objects.
    """
    directory = Path(directory)

    # Parse the structures
    structures = Structure.open_multi(directory / "optim.xyz")

    # Capture initialization stdout
    regex = r"""^(.*?)                # group-1 = everything before the banner
        (?=                       # look-ahead, do NOT consume banner
            ^-{55}\s*\r?\n        # 55 dashes
            0\ additional\ frames\ found\ in\ \S+\.xyz\s*\r?\n
            -{55}                 # 55 dashes
        )
    """
    match = re.search(
        regex,
        stdout,
        flags=re.MULTILINE | re.VERBOSE | re.DOTALL,
    )
    if not match:
        raise MatchNotFoundError(regex, stdout)
    initialization_stdout = match.group(1)

    # Capture the stdout for each gradient calculation
    regex = r"-=#=-\s+\(We'll Be Right Back\)\s+-=#=-\n(.*?)-=#=-\s+Now Returning to Optimizer\s+-=#=-"
    per_gradient_stdout = re.findall(regex, stdout, flags=re.DOTALL)
    if not per_gradient_stdout:
        raise MatchNotFoundError(regex, stdout)

    # Parse the gradient values from the stdout file
    from qccodec import decode

    # Create the trajectory
    trajectory: list[ProgramOutput] = []

    for structure, grad_stdout in zip(structures, per_gradient_stdout):
        # Create input data object for each structure and gradient in the trajectory.
        input_data_obj = ProgramInput(
            calctype=CalcType.gradient,
            structure=structure,
            model=input_data.model,
            keywords=input_data.keywords,
        )
        # Create the results object for each structure and gradient in the trajectory.
        full_grad_stdout = initialization_stdout + grad_stdout
        parsed_results = decode("terachem", CalcType.gradient, stdout=full_grad_stdout)
        assert isinstance(parsed_results, SinglePointResults)  # for mypy

        spr_data = parsed_results.model_dump()
        spr_data["energy"] = structure.extras[Structure._xyz_comment_key][0]
        results_obj = SinglePointResults(**spr_data)
        # Create the provenance object for each structure and gradient in the trajectory.
        prov = Provenance(
            program="terachem",
            program_version=parsed_results.extras["program_version"],
            scratch_dir=directory,
        )
        # Create the ProgramOutput object for each structure and gradient in the trajectory.
        traj_entry: ProgramOutput = ProgramOutput(
            input_data=input_data_obj,
            success=True,
            results=results_obj,
            provenance=prov,
            stdout=full_grad_stdout,
        )
        trajectory.append(traj_entry)

    return trajectory


def parse_version_control_details(contents: str) -> str:
    """Parse TeraChem git commit or Hg version from TeraChem stdout."""
    regex = r"(Git|Hg) Version: (\S*)"
    return re_search(regex, contents).group(2)


def parse_terachem_version(contents: str) -> str:
    """Parse TeraChem version from TeraChem stdout."""
    regex = r"TeraChem (v\S*)"
    return re_search(regex, contents).group(1)


def calculation_succeeded(contents: str) -> bool:
    """Determine from TeraChem stdout if a calculation completed successfully."""
    regex = r"Job finished:"
    if re.search(regex, contents):
        # If any match for a failure regex is found, the calculation failed
        return True
    return False


def parse_calctype(contents: str) -> CalcType:
    """Parse the calctype from TeraChem stdout."""
    calctypes = {
        r"SINGLE POINT ENERGY CALCULATIONS": CalcType.energy,
        r"SINGLE POINT GRADIENT CALCULATIONS": CalcType.gradient,
        r"FREQUENCY ANALYSIS": CalcType.hessian,
    }
    for regex, calctype in calctypes.items():
        match = re.search(regex, contents)
        if match:
            return calctype
    raise MatchNotFoundError(regex, contents)


@register(
    filetype=TeraChemFileType.STDOUT,
    calctypes=[CalcType.energy, CalcType.gradient],
    required=False,
    target=("extras", "excited_states"),
)
def parse_excited_states(contents: str) -> list[dict]:
    """Parse the excited state information from a TDDFT TeraChem stdout.

    Args:
        contents: The contents of the TeraChem TDDFT stdout file.
    Returns:
        A list of dictionaries containing the excited state information.
    Raises:
        MatchNotFoundError: If no excited state data is found.
    Notes:
        Converts the excitation energy from eV to Hartree.
    """
    regex = (
        r"^\s*(?:\d+)\s+(?P<energy>-?\d+\.\d+)\s+"
        r"(?P<exc_energy>-?\d+\.\d+)\s+"
        r"(?P<osc_strength>-?\d+\.\d+)\s+"
        r"(?P<s_squared>-?\d+\.\d+)\s+"
        r"(?P<max_ci_coeff>-?\d+\.\d+)\s+"
        r"(?P<excitation>\d+\s+->\s+\d+\s+:\s+\w+\s+->\s+\w+)$"
    )
    matches = re_finditer(regex, contents, re.MULTILINE)

    # Create a list of dictionaries for each excited state
    excited_states = []
    for match in matches:
        excited_state = match.groupdict()
        # Convert numeric values to floats
        for key, value in excited_state.items():
            if key != "excitation":  # Keep the excitation string as is
                excited_state[key] = float(value)

        # Convert the excitation energy to Hartree
        excited_state["exc_energy"] *= constants.EV_TO_HARTREE  # type: ignore
        excited_states.append(excited_state)

    return excited_states



"""Top level functions for the tcparse library"""

import logging
from importlib import import_module
from pathlib import Path
from typing import Any, Optional, Union

from qcio import (
    CalcType,
    ConformerSearchResults,
    OptimizationResults,
    ProgramInput,
    SinglePointResults,
    StructuredInputs,
    StructuredResults,
)

from .exceptions import DecoderError, EncoderError, MatchNotFoundError
from .models import (
    DataCollector,
    NativeInput,
)
from .parsers import *  # noqa: F403 Ensure all parsers get registered
from .registry import registry

__all__ = ["parse", "parse_results", "encode", "registry"]

logger = logging.getLogger(__name__)


RESULTS_TYPE_MAP = {
    CalcType.energy: SinglePointResults,
    CalcType.gradient: SinglePointResults,
    CalcType.hessian: SinglePointResults,
    CalcType.optimization: OptimizationResults,
    CalcType.transition_state: OptimizationResults,
    CalcType.conformer_search: ConformerSearchResults,
}


def decode(
    program: str,
    calctype: CalcType,
    *,
    stdout: Optional[str] = None,
    directory: Optional[Union[str, Path]] = None,
    input_data: Optional[StructuredInputs] = None,
    as_dict: bool = False,
) -> Union[StructuredResults, dict[str, Any]]:
    """Decode the output of a quantum chemistry program into a standardized output.
    
    Args:
        program: The QC program that generated the output file.
        calctype: The type of calculation that was run.
        stdout: The stdout file contents as a string.
        directory: The directory containing the output files.
        input_data: The input data used for the calculation.
            This is used to provide additional context for the parsers.
        as_dict: If True, return the results as a dictionary instead of a 
            StructuredResults object. Used mostly for testing purposes to enable 
            returning parsed data that isn't a fully valid StructuredResults object.

    Returns:
        A StructuredResults object containing the parsed data.

    Raises:
        DecoderError: If neither stdout nor directory is provided or if the program
            or calctype is not supported.
        MatchNotFoundError: If a required parser fails to find a match.
    """
    logger.info("Starting decode for program: %s with calctype: %s", program, calctype)
    if not stdout and not directory:
        raise ValueError("Either stdout, directory, or both must be provided.")

    # Import the program-specific module.
    try:
        mod = import_module(f"qccodec.parsers.{program}")
    except ImportError as e:
        logger.exception("Failed to import module qccodec.parsers.%s", program)
        raise DecoderError(f"No parsers found for program '{program}'.") from e

    # Create a generator for stdout (if provided) and all parsable files in directory
    files = mod.iter_files(stdout, directory)

    # Now iterate over the combined generator of all parsable files
    data_collector = DataCollector()
    for filetype, contents in files:
        # Look up the parsers for the given program, filetype, and calctype
        logger.debug("Processing file with filetype: %s", filetype)
        parser_specs = registry.get_parsers(program, filetype, calctype)
        logger.info("Found %d parser(s) for program '%s', filetype '%s', calctype '%s'", len(parser_specs), program, filetype, calctype) # noqa: E501
        
        for spec in parser_specs:
            logger.debug("Running parser '%s' for target '%s'", spec.parser.__name__, spec.target) # noqa: E501
            # Parse the contents using the parser
            try:
                if spec.filetype == "directory":
                    parsed_value: Any = spec.parser(directory, stdout, input_data)
                else:
                    parsed_value = spec.parser(contents)
                logger.info("Parser '%s' succeeded; returned value: %s", spec.parser.__name__, parsed_value) # noqa: E501
            # Raised if the parser can't find its data
            except MatchNotFoundError as e:
                if spec.required:
                    logger.error("Required parser '%s' failed; raising exception", spec.parser.__name__) # noqa: E501
                    raise
                else:
                    logger.info("Parser '%s' did not find a match but is not required.", spec.parser.__name__) # noqa: E501
            # Place the parsed value into the data collector
            else:
                # If the parser returns a dictionary, assign each key-value pair to the data collector
                if isinstance(parsed_value, dict):
                    for key, value in parsed_value.items():
                        data_collector.add_data(key, value)
                        logger.debug("Assigned parsed value to target '%s' on data_collector", (spec.target, key))
                # Otherwise, assign the parsed value to the specified target
                else:
                    assert spec.target is not None, "Target must be specified for non-dictionary parsed values." # for mypy
                    data_collector.add_data(spec.target, parsed_value)
                logger.debug("Assigned parsed value to target '%s' on data_collector", spec.target) # noqa: E501

    logger.info("Completed processing files; final data_collector state: %s", data_collector) # noqa: E501
    # Finally, construct and return the StructuredResults using the collected data.
    if as_dict:
        return dict(data_collector)
    return RESULTS_TYPE_MAP[calctype](**data_collector)

def encode(inp_data: ProgramInput, program: str) -> NativeInput:
    """Encode a ProgramInput object to a NativeInput object.

    Args:
        inp_data: The ProgramInput object to encode.
        program: The program for which to encode the input.

    Returns:
        A NativeInput object with the encoded input.

    Raises:
        EncoderError: If the calctype is not supported by the program's encoder or the
            input is invalid.
    """
    # Check that calctype is supported by the encoder
    encoder = import_module(f"qccodec.encoders.{program}")
    if inp_data.calctype not in encoder.SUPPORTED_CALCTYPES:
        raise EncoderError(f"Calctype '{inp_data.calctype}' not supported by encoder.")

    return encoder.encode(inp_data)



import json
from pathlib import Path

from qcio import ProgramOutput

# Load trajectory.json answer
traj_path = Path(__file__).parent / "trajectory.json"
traj_json = json.loads(traj_path.read_text())
trajectory: list[ProgramOutput] = [ProgramOutput(**item) for item in traj_json]

# Load excited-state-trajectory.json answer
es_traj_path = Path(__file__).parent / "excited-state-trajectory.json"
es_traj_json = json.loads(es_traj_path.read_text())
es_trajectory: list[ProgramOutput] = [ProgramOutput(**item) for item in es_traj_json]



import logging
import re

from qccodec.exceptions import MatchNotFoundError

logger = logging.getLogger(__name__)


def re_search(regex: str, contents: str, flags: int = 0) -> re.Match:
    """
    Search for a regex pattern in a string and return the first match.

    Raises:
        MatchNotFoundError if no match is found.
    """
    match = re.search(regex, contents, flags)
    if not match:
        raise MatchNotFoundError(regex, contents)
    return match


def re_finditer(regex: str, contents: str, flags: int = 0) -> list[re.Match]:
    """
    Search for a regex pattern in a string and return all match objects.

    Raises:
        MatchNotFoundError if no matches are found.
    """
    matches = list(re.finditer(regex, contents, flags))
    if not matches:
        raise MatchNotFoundError(regex, contents)
    return matches



from pathlib import Path

import pytest
from qcio import CalcType, ProgramOutput

from qccodec.exceptions import MatchNotFoundError
from qccodec.parsers.terachem import (
    calculation_succeeded,
    parse_calctype,
    parse_energy,
    parse_excited_states,
    parse_gradient,
    parse_hessian,
    parse_natoms,
    parse_nmo,
    parse_trajectory,
    parse_version,
    parse_version_control_details,
)

from .conftest import ParserTestCase, run_test_harness
from .data.terachem.answers import excited_states, gradients, hessians, trajectories

######################################################
##### Top level tests for all registered parsers #####
######################################################


test_cases = [
    ParserTestCase(
        name="Parse energy",
        parser=parse_energy,
        contents=Path("water.energy.out"),
        contents_stdout=True,
        calctype=CalcType.energy,
        success=True,
        answer=-76.3861099088,
    ),
    ParserTestCase(
        name="Parse energy no energy",
        parser=parse_energy,
        contents="No energy here",
        contents_stdout=True,
        calctype=CalcType.energy,
        success=False,
    ),
    ParserTestCase(
        name="Parse energy gradient",
        parser=parse_energy,
        contents=Path("water.gradient.out"),
        contents_stdout=True,
        calctype=CalcType.gradient,
        success=True,
        answer=-76.3861099088,
    ),
    ParserTestCase(
        name="Parse energy hessian",
        parser=parse_energy,
        contents=Path("water.frequencies.out"),
        contents_stdout=True,
        calctype=CalcType.hessian,
        success=True,
        answer=-76.3861099088,
    ),
    ParserTestCase(
        name="Parse energy excited states energy",
        parser=parse_energy,
        contents=Path("water.tddft.out"),
        contents_stdout=True,
        calctype=CalcType.energy,
        success=True,
        answer=-76.4002287974,
    ),
    ParserTestCase(
        name="Parse energy excited states positive",
        parser=parse_energy,
        contents="FINAL ENERGY: 124.38543379982 a.u",
        contents_stdout=True,
        calctype=CalcType.energy,
        success=True,
        answer=124.38543379982,
    ),
    ParserTestCase(
        name="Parse energy excited states negative integer",
        parser=parse_energy,
        contents="FINAL ENERGY: -7634 a.u",
        contents_stdout=True,
        calctype=CalcType.energy,
        success=True,
        answer=-7634,
    ),
    ParserTestCase(
        name="Parse energy excited states positive integer",
        parser=parse_energy,
        contents="FINAL ENERGY: 7123 a.u",
        contents_stdout=True,
        calctype=CalcType.energy,
        success=True,
        answer=7123,
    ),
    ParserTestCase(
        name="Parse version git",
        parser=parse_version,
        contents=Path("water.energy.out"),
        contents_stdout=True,
        calctype=CalcType.energy,
        success=True,
        answer="v1.9-2022.03-dev [4daa16dd21e78d64be5415f7663c3d7c2785203c]",
    ),
    ParserTestCase(
        name="Parse version hg",
        parser=parse_version,
        contents=Path("hg.out"),
        contents_stdout=True,
        calctype=CalcType.energy,
        success=True,
        answer="v1.5K [ccdev]",
    ),
    ParserTestCase(
        name="Parse water gradient",
        parser=parse_gradient,
        contents=Path("water.gradient.out"),
        contents_stdout=True,
        calctype=CalcType.gradient,
        success=True,
        answer=gradients.water,
    ),
    ParserTestCase(
        name="Parse caffeine gradient",
        parser=parse_gradient,
        contents=Path("caffeine.gradient.out"),
        contents_stdout=True,
        calctype=CalcType.gradient,
        success=True,
        answer=gradients.caffeine,
    ),
    ParserTestCase(
        name="Parse caffeine gradient from frequencies output",
        parser=parse_gradient,
        contents=Path("caffeine.frequencies.out"),
        contents_stdout=True,
        calctype=CalcType.hessian,
        success=True,
        answer=gradients.caffeine_frequencies,
    ),
    ParserTestCase(
        name="Parse water hessian",
        parser=parse_hessian,
        contents=Path("water.frequencies.out"),
        contents_stdout=True,
        calctype=CalcType.hessian,
        success=True,
        answer=hessians.water,
    ),
    ParserTestCase(
        name="Parse caffeine hessian",
        parser=parse_hessian,
        contents=Path("caffeine.frequencies.out"),
        contents_stdout=True,
        calctype=CalcType.hessian,
        success=True,
        answer=hessians.caffeine,
    ),
    ParserTestCase(
        name="Parse number of atoms water",
        parser=parse_natoms,
        contents=Path("water.energy.out"),
        contents_stdout=True,
        calctype=CalcType.energy,
        success=True,
        answer=3,
    ),
    ParserTestCase(
        name="Parse number of atoms caffeine",
        parser=parse_natoms,
        contents=Path("caffeine.gradient.out"),
        contents_stdout=True,
        calctype=CalcType.gradient,
        success=True,
        answer=24,
    ),
    ParserTestCase(
        name="Parse number of MOs water",
        parser=parse_nmo,
        contents=Path("water.energy.out"),
        contents_stdout=True,
        calctype=CalcType.energy,
        success=True,
        answer=13,
    ),
    ParserTestCase(
        name="Parse number of MOs caffeine",
        parser=parse_nmo,
        contents=Path("caffeine.gradient.out"),
        contents_stdout=True,
        calctype=CalcType.gradient,
        success=True,
        answer=146,
    ),
    ParserTestCase(
        name="Parse excited states water",
        parser=parse_excited_states,
        contents=Path("water.tddft.out"),
        contents_stdout=True,
        calctype=CalcType.energy,
        success=True,
        answer=excited_states.water,
    ),
    ParserTestCase(
        name="Parse excited states caffeine",
        parser=parse_excited_states,
        contents=Path("caffeine.tddft.out"),
        contents_stdout=True,
        calctype=CalcType.energy,
        success=True,
        answer=excited_states.caffeine,
    ),
    ParserTestCase(
        name="Parse excited states not found",
        parser=parse_excited_states,
        contents=Path("water.energy.out"),
        contents_stdout=True,
        calctype=CalcType.energy,
        decode_exc=False,
        success=False,
    ),
    ParserTestCase(
        name="Parse trajectory",
        parser=parse_trajectory,
        contents=Path("water.opt.out"),
        contents_stdout=True,
        calctype=CalcType.optimization,
        success=True,
        answer=trajectories.trajectory,
        clear_registry=False,
        extra_files=["optim.xyz"],
    ),
    ParserTestCase(
        name="Parse trajectory MatchNotFound",
        parser=parse_trajectory,
        contents=Path("water.energy.out"),
        contents_stdout=True,
        calctype=CalcType.optimization,
        success=False,
        clear_registry=False,
        extra_files=["optim.xyz"],
    ),
    ParserTestCase(
        name="Parse trajectory excited state",
        parser=parse_trajectory,
        contents=Path("excited-state-opt.out"),
        contents_stdout=True,
        calctype=CalcType.optimization,
        success=True,
        answer=trajectories.es_trajectory,
        clear_registry=False,
        extra_files=["es_optim.xyz"],
        extra_files_names=["optim.xyz"],
    ),
]


@pytest.mark.parametrize("test_case", test_cases, ids=lambda tc: tc.name)
def test_terachem_parsers(test_data_dir, prog_inp, tmp_path, test_case):
    """
    Tests the terachem parsers to ensure that they correctly parse the output files and
    behave correctly within the decode function.
    """
    # Water and excited state successful trajectories
    if test_case.success and "Parse trajectory" in test_case.name:
        # Update scratch_dir to the tmp_path for the trajectory test case
        for i in range(len(test_case.answer)):
            po_dict = test_case.answer[i].model_dump()
            po_dict["provenance"]["scratch_dir"] = tmp_path
            test_case.answer[i] = ProgramOutput(**po_dict)

    run_test_harness(test_data_dir, prog_inp, tmp_path, test_case)


####################################################
##### Test for non registered parser functions #####
####################################################


def test_parse_git_commit(terachem_file):
    contents = terachem_file("water.energy.out")
    assert (
        parse_version_control_details(contents)
        == "4daa16dd21e78d64be5415f7663c3d7c2785203c"  # pragma: allowlist secret
    )


def test_parse_excited_states_raises_exception_no_excited_states(terachem_file):
    """
    Tests the parse_excited_states function to ensure that it correctly raises
    an exception when no excited states are found in the output file.
    """
    contents = terachem_file("water.energy.out")

    with pytest.raises(MatchNotFoundError):
        parse_excited_states(contents)


###############################################
##### Unused parsers, but keeping for now #####
###############################################


@pytest.mark.parametrize(
    "filename,calctype",
    (
        ("water.energy.out", CalcType.energy),
        ("water.gradient.out", CalcType.gradient),
        ("water.frequencies.out", CalcType.hessian),
    ),
)
def test_parse_calctype(terachem_file, filename, calctype):
    contents = terachem_file(filename)
    assert parse_calctype(contents) == calctype


def test_parse_calctype_raises_exception():
    with pytest.raises(MatchNotFoundError):
        parse_calctype("No driver here")


def test_calculation_succeeded(terachem_file):
    contents = terachem_file("water.energy.out")
    assert calculation_succeeded(contents) is True
    assert (
        calculation_succeeded(
            """
        Incorrect purify value
        DIE called at line number 3572 in file terachem/params.cpp
         Job terminated: Thu Mar 23 03:47:12 2023
        """
        )
        is False
    )


@pytest.mark.parametrize(
    "filename,result",
    (
        ("failure.nocuda.out", False),
        ("failure.basis.out", False),
    ),
)
def test_calculation_succeeded_cuda_failure(terachem_file, filename, result):
    contents = terachem_file(filename)
    assert calculation_succeeded(contents) is result



import pytest
from qcio import ProgramInput
from qcio.utils import water

from qccodec.encoders.crest import _to_toml_dict, validate_input
from qccodec.exceptions import EncoderError


def test_validate_input(prog_inp):
    inp_obj = prog_inp("conformer_search")
    validate_input(inp_obj)

    with pytest.raises(EncoderError):
        inp_obj.keywords["charge"] = -1
        validate_input(inp_obj)

    with pytest.raises(EncoderError):
        inp_obj.keywords["uhf"] = 0
        validate_input(inp_obj)

    with pytest.raises(EncoderError):
        inp_obj.keywords["runtype"] = "imtd-gc"
        validate_input(inp_obj)


def test_toml_dict():
    """Test converting a ProgramInput object to a TOML dictionary for CREST."""

    weird_water = water.model_copy(update={"charge": -1, "multiplicity": 2})
    inp_obj = ProgramInput(
        structure=weird_water,
        calctype="conformer_search",
        model={"method": "gfn2"},
        keywords={"calculation": {"level": [{"alpb": "acetonitrile"}]}},
    )

    toml_dict = _to_toml_dict(inp_obj, "struct.xyz")

    assert toml_dict["input"] == "struct.xyz"
    assert toml_dict["runtype"] == "imtd-gc"
    assert toml_dict.get("threads") is not None  # added implicitly if not set

    # Adds values correctly to existing "calculation" key
    assert toml_dict["calculation"]["level"][0]["method"] == "gfn2"
    assert toml_dict["calculation"]["level"][0]["charge"] == -1
    assert toml_dict["calculation"]["level"][0]["uhf"] == 1
    assert toml_dict["calculation"]["level"][0]["alpb"] == "acetonitrile"

    # Respects explicitly set threads and handles no "calculation" key
    inp_obj = ProgramInput(
        structure=weird_water,
        calctype="conformer_search",
        model={"method": "gfn2"},
        keywords={"threads": 2},
    )

    toml_dict = _to_toml_dict(inp_obj, "struct.xyz")
    assert toml_dict["threads"] == 2
    assert toml_dict["calculation"]["level"][0]["method"] == "gfn2"
    assert toml_dict["calculation"]["level"][0]["charge"] == -1
    assert toml_dict["calculation"]["level"][0]["uhf"] == 1



import re
import subprocess
import sys
from datetime import datetime
from pathlib import Path

import toml


def get_repo_url():
    """Get the repository URL from pyproject.toml or ask the user for it."""
    try:
        with open("pyproject.toml") as file:
            pyproject = toml.load(file)
        repo_url = pyproject["project"]["urls"]["Source"]
        return repo_url
    except KeyError:
        return input("Enter the repository URL (e.g., https://github.com/user/repo): ")


def update_version_in_pyproject(version: str) -> None:
    """
    Update the version in pyproject.toml by replacing the line that sets the version
    in the [project] section with the new version string.
    
    This function uses a regex to find a line that starts with "version =", captures the surrounding quotes,
    and substitutes the new version.

    Args:
        version: The new version string (e.g. "0.7.6").
    """
    pyproject_path = Path("pyproject.toml")
    content = pyproject_path.read_text(encoding="utf-8")
    # This regex matches a line starting with 'version = "', then any characters until the next '"'
    new_content = re.sub(
        r'^(version\s=\s")[^"]*(")',
        r'\g<1>' + version + r'\g<2>',
        content,
        flags=re.MULTILINE,
    )
    pyproject_path.write_text(new_content, encoding="utf-8")


def update_changelog(version, repo_url):
    """Update the CHANGELOG.md file with the new version and today's date."""
    print("Updating CHANGELOG.md...")
    with open("CHANGELOG.md") as file:
        lines = file.readlines()

    today = datetime.today().strftime("%Y-%m-%d")
    new_entry = f"\n## [{version}] - {today}\n"

    # Insert the new entry after the ## [unreleased] section
    for i, line in enumerate(lines):
        if line.startswith("## [unreleased]"):
            lines.insert(i + 1, new_entry)
            break

    # Update links at the bottom
    unreleased_link = f"[unreleased]: {repo_url}/compare/{version}...HEAD\n"
    new_version_link = f"[{version}]: {repo_url}/releases/tag/{version}\n"

    # Ensure the new version link is added right after the [unreleased] link
    for i, line in enumerate(lines):
        if line.startswith("[unreleased]:"):
            lines[i] = unreleased_link
            lines.insert(i + 1, new_version_link)
            break

    with open("CHANGELOG.md", "w") as file:
        file.writelines(lines)


def run_git_commands(version):
    """Run the git commands to commit the changes, create a new tag, and push."""
    print("Running git commands...")
    subprocess.run(["git", "add", "."], check=True)
    subprocess.run(
        ["git", "commit", "-m", f"Bumped version to {version}. Updated CHANGELOG."],
        check=True,
    )
    subprocess.run(["git", "tag", version], check=True)
    subprocess.run(["git", "push", "--tags"], check=True)
    subprocess.run(["git", "push"], check=True)


def confirm_version(version: str):
    """Ask the user to confirm the version number before proceeding."""
    while True:
        response = (
            input(f"Are you sure you want to release {version}? (Y/N): ")
            .strip()
            .lower()
        )
        if response in ["y", "n"]:
            return response == "y"
        else:
            print("Invalid input. Please enter 'Y' or 'N'.")


def main():
    """Main entry point for the script."""
    from pathlib import Path

    if len(sys.argv) != 2:
        print("Usage: release.py <version>")
        sys.exit(1)

    version = sys.argv[1]

    original_pyproject = Path("pyproject.toml").read_text()
    original_changelog = Path("CHANGELOG.md").read_text()

    repo_url = get_repo_url()
    update_version_in_pyproject(version)
    update_changelog(version, repo_url)
    if confirm_version(version):
        print("Proceeding with the release...")
    else:
        print("Reverting changes...")
        Path("pyproject.toml").write_text(original_pyproject)
        Path("CHANGELOG.md").write_text(original_changelog)
        sys.exit(1)
    try:
        run_git_commands(version)
    except Exception:
        print("Reverting changes...")
        Path("pyproject.toml").write_text(original_pyproject)
        Path("CHANGELOG.md").write_text(original_changelog)
        sys.exit(1)


if __name__ == "__main__":
    main()



from qcio import CalcType, ProgramInput

from qccodec.exceptions import EncoderError
from qccodec.models import NativeInput

SUPPORTED_CALCTYPES = {
    CalcType.energy,
    CalcType.gradient,
    CalcType.hessian,
    CalcType.optimization,
    CalcType.transition_state,
}
XYZ_FILENAME = "geometry.xyz"
PADDING = 20  # padding between keyword and value in tc.in


def encode(inp_obj: ProgramInput) -> NativeInput:
    """Translate a ProgramInput into TeraChem inputs files.

    Args:
        inp_obj: The qcio ProgramInput object for a computation.

    Returns:
        NativeInput with .input being a tc.in file and .geometry an xyz file.
    """

    # calctype
    if inp_obj.calctype.value == CalcType.hessian:
        calctype = "frequencies"
    elif inp_obj.calctype.value == CalcType.optimization:
        calctype = "minimize"
        if not inp_obj.keywords.get("new_minimizer", "no") == "yes":
            raise EncoderError(
                "Only the new_minimizer is supported for optimizations. Add "
                "'new_minimizer': 'yes' to the keywords."
            )
    elif inp_obj.calctype.value == CalcType.transition_state:
        calctype = "ts"
    else:
        calctype = inp_obj.calctype.value

    # Collect lines for input file
    inp_lines = []
    inp_lines.append(f"{'run':<{PADDING}} {calctype}")
    # Structure
    inp_lines.append(f"{'coordinates':<{PADDING}} {XYZ_FILENAME}")
    inp_lines.append(f"{'charge':<{PADDING}} {inp_obj.structure.charge}")
    inp_lines.append(f"{'spinmult':<{PADDING}} {inp_obj.structure.multiplicity}")
    # Model
    inp_lines.append(f"{'method':<{PADDING}} {inp_obj.model.method}")
    inp_lines.append(f"{'basis':<{PADDING}} {inp_obj.model.basis}")

    # Keywords
    non_keywords = {
        "charge": ".structure.charge",
        "spinmult": ".structure.multiplicity",
        "run": ".calctype",
        "basis": ".model.basis",
        "method": ".model.method",
    }
    for key, value in inp_obj.keywords.items():
        # Check for keywords that should be passed as structured data
        if key in non_keywords:
            raise EncoderError(
                f"Keyword '{key}' should not be set as a keyword. It "
                f"should be set at '{non_keywords[key]}'",
            )
        # Lowercase booleans
        inp_lines.append(f"{key:<{PADDING}} {str(value).lower()}")
    return NativeInput(
        input_file="\n".join(inp_lines) + "\n",  # End file with newline
        geometry_file=inp_obj.structure.to_xyz(),
        geometry_filename=XYZ_FILENAME,
    )



"""Simple data models to support parsing of QM program output files."""

from dataclasses import dataclass
from typing import Any, Optional, Union

from .exceptions import DataCollectorError


class DataCollector(dict):
    """A dictionary for collecting data from parsers."""

    def add_data(self, target: Union[str, tuple[str, ...]], value: Any) -> None:
        """
        Assign a value into the DataCollector at the specified target.

        If target is a string, the value is assigned to that key.
        If target is a tuple, the method navigates through nested dictionaries,
        creating them as needed, and assigns the value at the final key.
        """
        keys = target if isinstance(target, tuple) else (target,)
        d = self
        for key in keys[:-1]:
            d = d.setdefault(key, {})
        if keys[-1] in d:
            raise DataCollectorError(
                f"Target '{keys}' already exists in DataCollector. You cannot add the same target twice."
            )
        d[keys[-1]] = value


@dataclass
class NativeInput:
    """Native input file data for a quantum chemistry program. 
    
    Writing these files to disk should produce a valid input.

    Attributes:
        input_file: input file for the program
        geometry: xyz file or other geometry file required for the calculation
        geometry_filename: filename of the geometry file referenced in the input
    """

    input_file: str
    geometry_file: Optional[str] = None
    geometry_filename: Optional[str] = None

    def __post_init__(self):
        """Ensure that geometry_filename is set if geometry is set."""
        if self.geometry_file and not self.geometry_filename:
            raise ValueError(
                "geometry_filename must be set if geometry_file is provided. "
                "Set geometry_filename to the name of the geometry file as referenced "
                "in the input file."
            )
